/* Copyright (c) 2007
   Colin Dewey (University of Wisconsin-Madison)
   cdewey@biostat.wisc.edu
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __DEWEYLAB_BIO_ALPHABET_AMINOACID_HH__
#define __DEWEYLAB_BIO_ALPHABET_AMINOACID_HH__

#include <string>
#include <vector>

#include "deweylab/bio/alphabet/Alphabet.hh"
#include "deweylab/util/charmanip.hh"

namespace deweylab {
namespace bio {
namespace alphabet {

class AminoAcid : public virtual Alphabet {
private:
	static const util::charmanip::Mapper hardMasker;
	static const util::charmanip::Mapper unMasker;
		
	const util::charmanip::Decoder decoder;
	const util::charmanip::Encoder encoder;

	const std::vector<std::string> names;
	const std::vector<std::string> short_names;
	
public:
	template<typename Iterator>
	AminoAcid(const std::string chars,
			  Iterator name_iter,
			  Iterator short_name_iter);
		
	encoded_char getSize() const;
		
	bool isMember(const char residue) const;

	encoded_char encode(const char residue) const;
	encoded_string encode(const std::string& s) const;
		
	char decode(const encoded_char c) const;
	std::string decode(const encoded_string& s) const;

	std::string getName(const char residue) const;
	std::string getShortName(const char residue) const;
	
	static void hardMaskInPlace(std::string& seq);
	static std::string hardMask(std::string seq);

	static void unMaskInPlace(std::string& seq);
	static std::string unMask(std::string seq);
};

template<typename Iterator>
AminoAcid::AminoAcid(const std::string chars,
					 Iterator name_iter,
					 Iterator short_name_iter) :
	decoder(chars),
	encoder(chars, false),
	names(name_iter, name_iter + chars.size()),
	short_names(short_name_iter, short_name_iter + chars.size())
{}

inline AminoAcid::encoded_char AminoAcid::getSize() const {
	return decoder.decoding.size();
}
		
inline bool AminoAcid::isMember(const char c) const {
	return encoder.isMember(c);
}
		
inline AminoAcid::encoded_char AminoAcid::encode(const char base) const {
	return encoder(base);
}

inline AminoAcid::encoded_string AminoAcid::encode(const std::string& s) const {
	return Alphabet::encode(s);
}

inline char AminoAcid::decode(const AminoAcid::encoded_char c) const {
	return decoder(c);
}

inline std::string AminoAcid::decode(const AminoAcid::encoded_string& s) const {
	return Alphabet::decode(s);
}

inline std::string AminoAcid::getName(const char base) const {
	return names[encode(base)];
}

const std::string PROTEIN_SHORT_NAMES[] =
{"Ala", "Cys", "Asp", "Glu", "Phe",
 "Gly", "His", "Ile", "Lys", "Leu",
 "Met", "Asn", "Pro", "Gln", "Arg",
 "Ser", "Thr", "Val", "Trp", "Tyr"};

const std::string PROTEIN_NAMES[] =
{"Alanine",
 "Cysteine",
 "Aspartic acid",
 "Glutamic acid",
 "Phenylalanine",
 "Glycine",
 "Histidine",
 "Isoleucine",
 "Lysine",
 "Leucine",
 "Methionine",
 "Asparagine",
 "Proline",
 "Glutamine",
 "Arginine",
 "Serine",
 "Threonine",
 "Valine",
 "Tryptophan",
 "Tyrosine"};

const AminoAcid PROTEIN("ACDEFGHIKLMNPQRSTVWY",
						PROTEIN_NAMES,
						PROTEIN_SHORT_NAMES);

} } }

#endif // __DEWEYLAB_BIO_ALPHABET_AMINOACID_HH__
