/* Copyright (c) 2007
   Colin Dewey (University of Wisconsin-Madison)
   cdewey@biostat.wisc.edu
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __DEWEYLAB_BIO_ALPHABET_NUCLEOTIDE_HH__
#define __DEWEYLAB_BIO_ALPHABET_NUCLEOTIDE_HH__

#include <string>
#include <vector>
#include <algorithm>

#include "deweylab/util/charmanip.hh"
#include "deweylab/bio/alphabet/Alphabet.hh"

namespace deweylab {
namespace bio {
namespace alphabet {

class Nucleotide : public virtual Alphabet {
private:
	static const util::charmanip::Mapper hardMasker;
	static const util::charmanip::Mapper unMasker;
		
	static const util::charmanip::Mapper transcriber;
	static const util::charmanip::Mapper reverseTranscriber;

	static const util::charmanip::Encoder nibEncoder;
	static const util::charmanip::Decoder nibDecoder;
	static const util::charmanip::Decoder nibHardMaskDecoder;
	static const util::charmanip::Decoder nibUnmaskDecoder;
		
	const util::charmanip::Decoder decoder;
	const util::charmanip::Encoder encoder;
	const util::charmanip::Mapper complementer;
	const std::vector<std::string> names;

public:
	template<typename Iterator>
	Nucleotide(const std::string chars,
			   const std::string complement,
			   Iterator name_iter);
		
	char complement(const char base) const;
		
	encoded_char getSize() const;
		
	bool isMember(const char c) const;
		
	encoded_char encode(const char base) const;
	encoded_string encode(const std::string& s) const;
		
	char decode(const encoded_char c) const;
	std::string decode(const encoded_string& s) const;

	std::string getName(const char c) const;
		
	static std::string nibEncode(const std::string& seq);
	static std::string nibDecode(const std::string& seq,
								 bool hardMask=false,
								 bool unmask=false);

	void complementInPlace(std::string& seq) const;
	std::string complement(std::string seq) const;
		
	void reverseComplementInPlace(std::string& seq) const;
	std::string reverseComplement(std::string seq) const;

	static void transcribeInPlace(std::string& seq);
	static std::string transcribe(std::string seq);

	static void reverseTranscribeInPlace(std::string& seq);
	static std::string reverseTranscribe(std::string seq);

	static void hardMaskInPlace(std::string& seq);
	static std::string hardMask(std::string seq);

	static void unMaskInPlace(std::string& seq);
	static std::string unMask(std::string seq);
};

template<typename Iterator>
Nucleotide::Nucleotide(const std::string chars,
					   const std::string complement,
					   Iterator name_iter) :
	decoder(chars),
	encoder(chars, false),
	complementer(chars, complement, false),
	names(name_iter, name_iter + chars.size())
{}

inline char Nucleotide::complement(const char base) const {
	return complementer(base);
}
		
inline Nucleotide::encoded_char Nucleotide::getSize() const {
	return decoder.decoding.size();
}
		
inline bool Nucleotide::isMember(const char c) const {
	return encoder.isMember(c);
}
		
inline Nucleotide::encoded_char Nucleotide::encode(const char base) const {
	return encoder(base);
}

inline Nucleotide::encoded_string Nucleotide::encode(const std::string& s) const {
	return Alphabet::encode(s);
}

inline char Nucleotide::decode(const Nucleotide::encoded_char c) const {
	return decoder(c);
}

inline std::string Nucleotide::decode(const Nucleotide::encoded_string& s) const {
	return Alphabet::decode(s);
}

inline std::string Nucleotide::getName(const char base) const {
	return names[encode(base)];
}

const std::string DNA_NAMES[] = {"Adenine", "Cytosine", "Guanine", "Thymine"};
const std::string DNA_PURINES_NAMES[] = {"Adenine", "Guanine"};
const std::string DNA_PYRIMIDINES_NAMES[] = {"Cytosine", "Thymine"};

const std::string RNA_NAMES[] = {"Adenine", "Cytosine", "Guanine", "Uracil"};
const std::string RNA_PURINES_NAMES[] = {"Adenine", "Guanine"};
const std::string RNA_PYRIMIDINES_NAMES[] = {"Cytosine", "Uracil"};

const Nucleotide DNA("ACGT", "TGCA", DNA_NAMES);
const Nucleotide DNA_PURINES("AG", "TC", DNA_PURINES_NAMES);
const Nucleotide DNA_PYRIMIDINES("CT", "GA", DNA_PYRIMIDINES_NAMES);

const Nucleotide RNA("ACGU", "UGCA", RNA_NAMES);
const Nucleotide RNA_PURINES("AG", "UC", RNA_PURINES_NAMES);
const Nucleotide RNA_PYRIMIDINES("CU", "GA", RNA_PYRIMIDINES_NAMES);

} } }

#endif // __DEWEYLAB_BIO_ALPHABET_NUCLEOTIDE_HH__
