/* Copyright (c) 2007
   Colin Dewey (University of Wisconsin-Madison)
   cdewey@biostat.wisc.edu
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "deweylab/bio/formats/agp/Record.hh"

#include <stdexcept>

namespace deweylab {
namespace bio {
namespace formats {
namespace agp {

	// Default constructor
	Record::Record() {
	}
	
	// Source sequence constructor
	Record::Record(const std::string& object,
				   const genome::Position objectStart,
				   const genome::Position objectEnd,
				   const size_t partNumber,
				   const char componentType,
				   const std::string& componentID,
				   const genome::Position componentStart,
				   const genome::Position componentEnd,
				   const genome::Strand componentStrand) :
		object(object),
		objectStart(objectStart),
		objectEnd(objectEnd),
		partNumber(partNumber),
		componentType(componentType),
		componentID(componentID),
		componentStart(componentStart),
		componentEnd(componentEnd),
		componentStrand(componentStrand) {
		if (componentType == 'N') {
			throw std::runtime_error("Attempted construction of non-gap AGP "
									 "record with source type = N");
		}
	}
		
	// Gap constructor
	Record::Record(const std::string& object,
				   const genome::Position objectStart,
				   const genome::Position objectEnd,
				   const size_t partNumber,
				   const char componentType,
				   const genome::Distance gapLength,
				   const std::string& gapType,
				   const bool isLinkage) :
		object(object),
		objectStart(objectStart),
		objectEnd(objectEnd),
		partNumber(partNumber),
		componentType(componentType),
		gapLength(gapLength),
		gapType(gapType),
		isLinkage(isLinkage) {
		if (componentType != 'N') {
			throw std::runtime_error("Attempted construction of gap AGP "
									 "record with source type != N");
		}
	}
	
bool Record::isGap() const {
	return componentType == GAP;
}

// Interval interface
std::string Record::getChrom() const { return object; }
genome::Position Record::getStart() const { return objectStart - 1; }
genome::Position Record::getEnd() const { return objectEnd; }
genome::Strand Record::getStrand() const { return '+'; }
void Record::setChrom(const std::string& chrom) {
	object = chrom;
}
void Record::setStart(const genome::Position start) {
	objectStart = start + 1;
}
void Record::setEnd(const genome::Position end) {
	objectEnd = end;
}
void Record::setStrand(const genome::Strand strand) {
	throw std::logic_error("Can not set the strand of the object in an AGP "
						   "record");
}
	
// Interval map interface
std::string Record::getSourceChrom() const { return componentID; }
genome::Position Record::getSourceStart() const { return componentStart - 1; }
genome::Position Record::getSourceEnd() const { return componentEnd; }
genome::Strand Record::getSourceStrand() const { return componentStrand; }
	
std::string Record::getTargetChrom() const { return object; }
genome::Position Record::getTargetStart() const { return objectStart - 1; }
genome::Position Record::getTargetEnd() const { return objectEnd; }
genome::Strand Record::getTargetStrand() const { return '+'; }

void Record::setSourceChrom(const std::string& chrom) {
	componentID = chrom;
}
	
void Record::setSourceStart(const genome::Position start) {
	componentStart = start + 1;
}
	
void Record::setSourceEnd(const genome::Position end) {
	componentEnd = end;
}
	
void Record::setSourceStrand(const genome::Strand strand) {
	componentStrand = strand;
}
	
void Record::setTargetChrom(const std::string& chrom) {
	object = chrom;
}
	
void Record::setTargetStart(const genome::Position start) {
	objectStart = start + 1;
}
	
void Record::setTargetEnd(const genome::Position end) {
	objectEnd = end;
}
	
void Record::setTargetStrand(const genome::Strand strand) {
	setStrand(strand);
}
	
} } } }
