/* Copyright (c) 2007
   Colin Dewey (University of Wisconsin-Madison)
   cdewey@biostat.wisc.edu
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __DEWEYLAB_BIO_FORMATS_FASTQ_QUALITYSCOREENCODING_HH__
#define __DEWEYLAB_BIO_FORMATS_FASTQ_QUALITYSCOREENCODING_HH__

#include <cmath>
#include <vector>
#include <string>
#include <limits>

namespace deweylab {
namespace bio {
namespace formats {
namespace fastq {

typedef unsigned char PhredScore;
typedef signed char SolexaScore;

class QualityScoreEncoding {
public:
    virtual PhredScore decode_phred(char c) const = 0;
    virtual SolexaScore decode_solexa(char c) const = 0;

    virtual char encode_phred(PhredScore s) const = 0;
    virtual char encode_solexa(SolexaScore s) const = 0;
    virtual char encode(char c,
                        const QualityScoreEncoding* source_encoding) const = 0;
    virtual std::string encode(const std::string& s,
                               const QualityScoreEncoding* source_encoding) const = 0; 
    
    virtual char min_encoded_val() const = 0;
    virtual char max_encoded_val() const = 0;

    virtual ~QualityScoreEncoding() {}
};

SolexaScore phred_to_solexa(PhredScore s);
PhredScore solexa_to_phred(SolexaScore s);

template<PhredScore MIN_PHRED_ENCODABLE,
         PhredScore MAX_PHRED_ENCODABLE,
         PhredScore ENCODING_OFFSET>
class PhredEncoding : public QualityScoreEncoding {
public:
    PhredEncoding()
        : solexa_scores_(MAX_PHRED_ENCODABLE -
                         MIN_PHRED_ENCODABLE + 1),
          encoded_solexa_(std::numeric_limits<SolexaScore>::max() -
                          std::numeric_limits<SolexaScore>::min() + 1) {
        for (size_t i = 0; i < solexa_scores_.size(); ++i) {
            solexa_scores_[i] = phred_to_solexa(i + MIN_PHRED_ENCODABLE);
        }
        SolexaScore s = std::numeric_limits<SolexaScore>::min();
        for (size_t i = 0; i < encoded_solexa_.size(); ++i, ++s) {
            encoded_solexa_[i] = encode_phred(solexa_to_phred(s));
        }
    }

    static const PhredEncoding& instance() {
        static const PhredEncoding singleton;
        return singleton;
    }
    
    PhredScore decode_phred(char c) const {
        return static_cast<PhredScore>(c) - ENCODING_OFFSET;
    }
    
    SolexaScore decode_solexa(char c) const {
        return solexa_scores_[c - min_encoded_val()];
    }
    
    char encode_phred(PhredScore s) const {
        return std::min(std::max(s, MIN_PHRED_ENCODABLE),
                        MAX_PHRED_ENCODABLE) + ENCODING_OFFSET;
    }

    char encode_solexa(SolexaScore s) const {
        return encoded_solexa_[s - std::numeric_limits<SolexaScore>::min()];
    }

    char encode(char c,
                const QualityScoreEncoding* source_encoding) const {
        if (this == source_encoding) {
            return c;
        } else {
            return encode_phred(source_encoding->decode_phred(c));
        }
    }
    
    std::string encode(const std::string& s,
                       const QualityScoreEncoding* source_encoding) const {
        if (this == source_encoding) {
            return s;
        } else {
            std::string encoded = s;
            for (size_t i = 0; i < encoded.size(); ++i) {
                encoded[i] = encode_phred(source_encoding->decode_phred(encoded[i]));
            }
            return encoded;
        }
    }
    
    char min_encoded_val() const {
        return MIN_PHRED_ENCODABLE + ENCODING_OFFSET;
    }

    char max_encoded_val() const {
        return MAX_PHRED_ENCODABLE + ENCODING_OFFSET;
    }

private:
    std::vector<SolexaScore> solexa_scores_;
    std::vector<char> encoded_solexa_;
};

template<SolexaScore MIN_SOLEXA_ENCODABLE,
         SolexaScore MAX_SOLEXA_ENCODABLE,
         SolexaScore ENCODING_OFFSET>
class SolexaEncoding : public QualityScoreEncoding {
public:
    SolexaEncoding()
        : phred_scores_(MAX_SOLEXA_ENCODABLE - MIN_SOLEXA_ENCODABLE + 1),
          encoded_phred_(std::numeric_limits<PhredScore>::max() -
                         std::numeric_limits<PhredScore>::min() + 1) {
        for (size_t i = 0; i < phred_scores_.size(); ++i) {
            phred_scores_[i] = solexa_to_phred(i + MIN_SOLEXA_ENCODABLE);
        }
        PhredScore s = std::numeric_limits<PhredScore>::min();
        for (size_t i = 0; i < encoded_phred_.size(); ++i, ++s) {
            encoded_phred_[i] = encode_solexa(phred_to_solexa(s));
        }
    }

    static const SolexaEncoding& instance() {
        static const SolexaEncoding singleton;
        return singleton;
    }
    
    PhredScore decode_phred(char c) const {
        return phred_scores_[c - min_encoded_val()];
    }
    
    SolexaScore decode_solexa(char c) const {
        return static_cast<SolexaScore>(c) - ENCODING_OFFSET;
    }

    char encode_phred(PhredScore s) const {
        return encoded_phred_[s - std::numeric_limits<PhredScore>::min()];
    }
    
    char encode_solexa(SolexaScore s) const {
        return std::min(std::max(s, MIN_SOLEXA_ENCODABLE),
                        MAX_SOLEXA_ENCODABLE) + ENCODING_OFFSET;
    }

    char encode(char c,
                const QualityScoreEncoding* source_encoding) const {
        if (this == source_encoding) {
            return c;
        } else {
            return encode_solexa(source_encoding->decode_solexa(c));
        }
    }
    
    std::string encode(const std::string& s,
                       const QualityScoreEncoding* source_encoding) const {
        if (this == source_encoding) {
            return s;
        } else {
            std::string encoded = s;
            for (size_t i = 0; i < encoded.size(); ++i) {
                encoded[i] = encode_solexa(source_encoding->decode_solexa(encoded[i]));
            }
            return encoded;
        }
    }
    
    char min_encoded_val() const {
        return MIN_SOLEXA_ENCODABLE + ENCODING_OFFSET;
    }

    char max_encoded_val() const {
        return MAX_SOLEXA_ENCODABLE + ENCODING_OFFSET;
    }
    
private:
    std::vector<PhredScore> phred_scores_;
    std::vector<char> encoded_phred_;
};

typedef PhredEncoding<0, 93, 33> SangerEncoding;
typedef PhredEncoding<0, 62, 64> IlluminaEncoding;
typedef SolexaEncoding<-5, 62, 64> OldIlluminaEncoding;

const QualityScoreEncoding* get_quality_score_encoding(const std::string& name);

typedef std::vector<std::string> QualityScoreNameList;
const QualityScoreNameList& get_quality_score_names();

}
}
}
}

#endif // __DEWEYLAB_BIO_FORMATS_FASTQ_QUALITYSCOREENCODING_HH__
