/* Copyright (c) 2007
   Colin Dewey (University of Wisconsin-Madison)
   cdewey@biostat.wisc.edu
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __DEWEYLAB_BIO_FORMATS_GFF_RECORD_HH__
#define __DEWEYLAB_BIO_FORMATS_GFF_RECORD_HH__

#include <string>
#include <iosfwd>
#include <list>

#include "deweylab/bio/genome/BasicInterval.hh"
#include "deweylab/util/io.hh"
#include "boost/unordered_map.hpp"

namespace deweylab {
namespace bio {
namespace formats {
namespace gff {

class Record {
public:
    typedef float Score;

    struct Attribute {
		std::string tag;
		std::list<std::string> values;
    };
		
    Record();
		
    const std::string& getSeqname() const;
    const std::string& getSource() const;
    const std::string& getFeature() const;
    genome::Position getStart() const;
    genome::Position getEnd() const;
    Score getScore() const;
    genome::Strand getStrand() const;
    unsigned char getFrame() const;
    genome::BasicInterval getInterval() const;

    bool hasScore() const;
    bool hasStrand() const;
    bool hasFrame() const;

    void setInterval(const genome::Interval& i);
    void setSeqname(const std::string& seqname);
    void setSource(const std::string& source);
    void setFeature(const std::string& feature);
    void setStart(const genome::Position start);
    void setEnd(const genome::Position end);
    void setScore(const Score score);
    void setStrand(const genome::Strand strand);
    void setFrame(const unsigned char frame);

    void unsetScore();
    void unsetStrand();
    void unsetFrame();

    Attribute& getAttribute(const std::string& tag) const;
    bool hasAttribute(const std::string& tag) const;
    Attribute& addAttribute(const std::string& tag);
    void removeAttribute(const std::string& tag);

    friend std::ostream& operator<<(std::ostream& strm, const Record& r);
    friend std::istream& operator>>(std::istream& strm, Record& r);
    friend void operator>>(const std::string& tag, Record& r);

    bool operator<(const Record& r) const;
		
private:
    typedef boost::unordered_map<std::string, const std::string*> StringMap;
    typedef std::list<Attribute> AttributeList;

    struct AttributeFinder {
		const std::string& tag;
		AttributeFinder(const std::string& tag);
		bool operator()(const Attribute& a) const;
    };

    const std::string* storeString(const std::string& s,
								   StringMap& strmap);
    void parseAttributes() const;
		
    static const std::string THE_EMPTY_STRING;
    static StringMap seqnames;
    static StringMap sources;
    static StringMap features;
		
    const std::string* seqname;
    const std::string* source;
    const std::string* feature;
    genome::Position start;
    genome::Position end;
    float score;
    char strand;
    char frame;
    mutable std::string attributeString;
    mutable AttributeList attributes;
		
    bool validScore;
};

inline Record::AttributeFinder::AttributeFinder(const std::string& tag)
    : tag(tag) {}
inline bool Record::AttributeFinder::operator()(const Attribute& a) const {
    return a.tag == tag;
}
	
inline const std::string& Record::getSeqname() const { return *seqname; }
inline const std::string& Record::getSource() const { return *source; }
inline const std::string& Record::getFeature() const { return *feature; }
inline genome::Position Record::getStart() const { return start; }
inline genome::Position Record::getEnd() const { return end; }
inline Record::Score Record::getScore() const { return score; }
inline genome::Strand Record::getStrand() const { return strand; }
inline unsigned char Record::getFrame() const { return frame - '0'; }

inline bool Record::hasScore() const { return validScore; }
inline bool Record::hasStrand() const { return strand != '.'; }
inline bool Record::hasFrame() const { return frame != '.'; }
	
inline void Record::setSeqname(const std::string& seqname) {
    this->seqname = storeString(seqname, seqnames);
}
inline void Record::setSource(const std::string& source) {
    this->source = storeString(source, sources);
}
inline void Record::setFeature(const std::string& feature) {
    this->feature = storeString(feature, features);
}

inline void Record::setStart(const genome::Position start) {
    this->start = start;
}
inline void Record::setEnd(const genome::Position end) {
    this->end = end;
}

inline void Record::setScore(const Score score) {
    this->score = score;
    this->validScore = true;
}
	
inline void Record::setStrand(const genome::Strand strand) {
    this->strand = strand;
}
inline void Record::setFrame(const unsigned char frame) {
    this->frame = '0' + frame;
}

inline void Record::unsetScore() { this->validScore = false; }
inline void Record::unsetStrand() { this->strand = '.'; }
inline void Record::unsetFrame() { this->frame = '.'; }

} } } }

#endif // __DEWEYLAB_BIO_FORMATS_GFF_RECORD_HH__
