/* Copyright (c) 2007
   Colin Dewey (University of Wisconsin-Madison)
   cdewey@biostat.wisc.edu
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __DEWEYLAB_BIO_FORMATS_SDB_DB_HH__
#define __DEWEYLAB_BIO_FORMATS_SDB_DB_HH__

#include <stdexcept>
#include <fstream>

#include "boost/iterator/indirect_iterator.hpp"

#include "deweylab/bio/formats/sdb/Record.hh"
#include "deweylab/bio/genome/Interval.hh"
#include "deweylab/bio/alphabet/AmbiguousNucleotide.hh"
#include "deweylab/filesystem/Path.hh"

namespace deweylab {
namespace bio {
namespace formats {
namespace sdb {

class NotFoundError : public std::runtime_error {
public:
    NotFoundError(const std::string& what);
};

class OutOfBoundsError : public std::runtime_error {
public:
    OutOfBoundsError(const std::string& what);
};
	
inline NotFoundError::NotFoundError(const std::string& what)
    : std::runtime_error(what) {
}

inline OutOfBoundsError::OutOfBoundsError(const std::string& what)
    : std::runtime_error(what) {
}

class DB {
public:
    typedef boost::indirect_iterator<std::vector<Record*>::iterator> Iterator;

    DB(bool cache=false);
    ~DB();

    void open(const filesystem::Path& filename,
			  bool readonly=true,
			  bool create=false);

    void close();

    void putRec(const std::string& title,
				const std::string& sequence,
				const bool compressed=false);
		
    void getRec(const std::string& title, Record& rec);
    void getRec(const uint32_t recNum, Record& rec);

    std::string getSeq(const std::string& title);
    std::string getSeq(const std::string& title,
					   const uint32_t start,
					   const uint32_t end,
					   const char strand='+',
					   const alphabet::Nucleotide& alphabet=alphabet::AMBIGUOUS_DNA);

    std::string getSeq(const genome::Interval& i,
					   const alphabet::Nucleotide& alphabet=alphabet::AMBIGUOUS_DNA);

    std::string getSeq(const uint32_t recNum);
    std::string getSeq(const uint32_t recNum,
					   const uint32_t start,
					   const uint32_t end,
					   const char strand='+',
					   const alphabet::Nucleotide& alphabet=alphabet::AMBIGUOUS_DNA);
		
    Iterator begin();
    Iterator end();

    void setCache(bool cache=true);
    void readIndex();

private:
    void openFile(const filesystem::Path& filename,
				  const char* openmode);
    void openForNew(const filesystem::Path& filename);
    void openForAppend(const filesystem::Path& filename);
    void openForRead(const filesystem::Path& filename);

    void readHeader();
    void writeHeader();
    uint32_t headerSize() const;

    uint32_t calcSizes(std::vector<Record*>::iterator start,
						   std::vector<Record*>::iterator end);
    void writeRecs(std::vector<Record*>::iterator start,
				   std::vector<Record*>::iterator end);

    void writeIndex();
    void readIndexRecursive();
    void sortIndex();

    bool isIndexRead() const;
    bool isIndexSorted() const;

    void cacheRec(const std::string& title);
    void cacheRec(const uint32_t recNum);
		
    bool lookupRec(const std::string& title);
    bool lookupRec(const uint32_t recNum);
		
    bool lookupRec(const std::string& title,
				   std::vector<Record*>::iterator begin,
				   std::vector<Record*>::iterator end);
    bool lookupRec(const uint32_t recNum,
				   std::vector<Record*>::iterator begin,
				   std::vector<Record*>::iterator end);

    bool lookupRecInBuffer(const std::string& title,
						   std::istream& buffer);
		
    void reset();

    bool cache;
    Record lastRecord;
		
    bool opened;
    bool readonly;
    bool indexSorted;
    FILE* strm;
    uint32_t version;
    uint32_t numRecs;
    off_t seqSize;
    uint32_t indexSize;

    std::vector<Record*> recs;
	
    static const uint32_t SDB_MAGIC_NUMBER;
    static const uint32_t SDB_VERSION;

    friend class Record;
};
			
} } } }

#endif // __DEWEYLAB_BIO_FORMATS_SDB_DB_HH__
