/* Copyright (c) 2007
   Colin Dewey (University of Wisconsin-Madison)
   cdewey@biostat.wisc.edu
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "deweylab/bio/phylogenetic/Tree.hh"
#include "deweylab/bio/phylogenetic/TreeVisitor.hh"

#include <iostream>
#include <queue>
#include <limits>

namespace deweylab {
namespace bio {
namespace phylogenetic {

Tree::Tree(const std::string& label, double branchLength)
	: label(label),
	  num(0),
	  branchLength(branchLength),
	  children() {
}

TreePtr Tree::copy() const {
	TreePtr t(new Tree(label, branchLength));
	for (size_t i = 0; i < children.size(); ++i) {
		t->addChild(children[i]->copy());
	}
	return t;
}
	
TreePtr Tree::getSubtree(boost::unordered_set<std::string>& taxa,
						 bool keep_children) const {
	bool includedTaxon = (taxa.find(label) != taxa.end());
	
	if (keep_children and includedTaxon) {
		return copy();
	}

	ChildList subtreeChildren;
	for (size_t i = 0; i < children.size(); ++i) {
		TreePtr t = children[i]->getSubtree(taxa, keep_children);
		if (t != NULL) {
			subtreeChildren.push_back(t);
		}
	}
	
	if (includedTaxon or subtreeChildren.size() > 1) {
		TreePtr t(new Tree(label, branchLength));
		for (size_t i = 0; i < subtreeChildren.size(); ++i) {
			t->addChild(subtreeChildren[i]);
		}
		return t;
	} else if (subtreeChildren.empty()) {
		return TreePtr();
	} else {
		TreePtr child = subtreeChildren.front();
		child->setBranchLength(child->getBranchLength() + branchLength);
		return child;
	}
}
	
class TreeNumbererVisitor : public TreeVisitor {
private:
	unsigned int counter;
public:
	TreeNumbererVisitor() : counter(0) {}
	void postVisitNode(const TreePtr& t) { t->setNum(counter++); }
};

void set_numbers(TreePtr root) {
	TreeNumbererVisitor numberer;
	numberer.dfsTraverse(root);
}

size_t Tree::getNumLeaves() const {
	if (isLeaf()) {
		return 1;
	} else {
		size_t numLeaves = 0;
		ChildList::const_iterator it;
		for (it = children.begin(); it != children.end(); ++it) {
			numLeaves += (*it)->getNumLeaves();
		}
		return numLeaves;
	}
}
		
size_t Tree::getNumNodes() const {
	size_t numNodes = 1;
	ChildList::const_iterator it;
	for (it = children.begin(); it != children.end(); ++it) {
		numNodes += (*it)->getNumNodes();
	}
	return numNodes;
}


class TreePairwiseDistanceVisitor : public ConstTreeVisitor {
private:
	util::Matrix<double>& m;
public:
	TreePairwiseDistanceVisitor(util::Matrix<double>& m)
		: m(m)
		{
			for (size_t i = 0; i < m.getNumRows(); ++i) {
				for (size_t j = 0; j < m.getNumCols(); ++j) {
					m(i, j) = (i == j ? 0.0 : std::numeric_limits<double>::min());
				}
			}
		}

	void preVisitTree(const ConstTreePtr& tree) {
		size_t num_nodes = tree->getNumNodes();
		m.resize(num_nodes, num_nodes);
	}
	
	void preVisitEdge(const ConstTreePtr& parent, const ConstTreePtr& child) {
		for (size_t i = 0; i < m.getNumCols(); ++i) {
			if (i == child->getNum()) {
				continue;
			}
			double parentDistance = m(parent->getNum(), i);
			if (parentDistance != std::numeric_limits<double>::min()) {
				double childDistance = parentDistance + child->getBranchLength();
				m(child->getNum(), i) = childDistance;
				m(i, child->getNum()) = childDistance;
			}
		}
	}
};
	
void get_pairwise_distances(ConstTreePtr root, util::Matrix<double>& distances) {
	TreePairwiseDistanceVisitor visitor(distances);
	visitor.dfsTraverse(root);
}

class TreeOrdererVisitor : public ConstTreeVisitor {
private:
	std::vector<ConstTreePtr>& order;
public:
	TreeOrdererVisitor(std::vector<ConstTreePtr>& order) : order(order) {}
	void postVisitNode(const ConstTreePtr& t) { order.push_back(t); }
};

void get_topological_order(ConstTreePtr root, std::vector<ConstTreePtr>& order) {
	TreeOrdererVisitor visitor(order);
	visitor.dfsTraverse(root);
}

} } }
