/* Copyright (c) 2007
   Colin Dewey (University of Wisconsin-Madison)
   cdewey@biostat.wisc.edu
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __DEWEYLAB_BIO_PHYLOGENETIC_TREE_HH__
#define __DEWEYLAB_BIO_PHYLOGENETIC_TREE_HH__

#include <string>
#include <vector>
#include "boost/shared_ptr.hpp"
#include "boost/unordered_set.hpp"

#include "deweylab/util/matrix.hh"
		
namespace deweylab {
namespace bio {
namespace phylogenetic {

class Tree;
typedef boost::shared_ptr<Tree>	TreePtr;
typedef boost::shared_ptr<const Tree> ConstTreePtr;
typedef std::vector<TreePtr> ChildList;

// A node in a phylogenetic tree
class Tree {
private:
	std::string label;
	size_t num;
	double branchLength;
	ChildList children;

public:
	
	// Construct from optional node label and optional branch
	// length from parent node
	Tree(const std::string& label = "", double branchLength = 0.0);

	// Returns true if this node is a leaf node
	bool isLeaf() const { return children.empty(); }

	// Add tree D as a descendant of this tree node
	void addChild(TreePtr d) { children.push_back(d); }

	// Get the label of this node
	const std::string& getLabel() const { return label; }

	// Get the length of the branch connecting this node to its parent
	double getBranchLength() const { return branchLength; }

	// Get the topological number of this node.  The method
	// setNumbers must have been called previously for this number
	// to be valid.
	size_t getNum() const { return num; }

	// Get the number of nodes in this (sub)tree
	size_t getNumNodes() const;

	// Get the number of leaf nodes in this (sub)tree
	size_t getNumLeaves() const;

	// Get the number of children of this tree node
	size_t getNumChildren() const { return children.size(); }

	// Get the ith descendant
	TreePtr getChild(size_t i) const { return children.at(i); }
		
	// Get the children of this tree node
	const ChildList& getChildren() const { return children; }
	
	// Remove the ith descendant
	void removeChild(size_t i) {
		children.erase(children.begin() + i);
	}
	
	// Remove all children of this tree node
	void removeAllChildren() { children.clear(); }

	// Set the label of this node
	void setLabel(const std::string& l) { label = l; }

	// Set the length of the branch connecting this node to its parent
	void setBranchLength(double bl) { branchLength = bl; }

	// Set the topological number of this node.  Used by
	// setNumbers method
	void setNum(size_t n) { num = n; }

	TreePtr copy() const;
		
	// Return tree with given TAXA contained within this tree
	TreePtr getSubtree(boost::unordered_set<std::string>& taxa,
					   bool keep_children = false) const;
};

// Set the topological numbers of the nodes in a tree
void set_numbers(TreePtr tree);

// Get the pairwise distances between all nodes in the matrix
// DISTANCES
void get_pairwise_distances(ConstTreePtr tree, util::Matrix<double>& distances);

// Get all of the nodes in this tree in toplogical order in
// the vector ORDER
void get_topological_order(ConstTreePtr tree, std::vector<ConstTreePtr>& order);

} }

};
#endif // __DEWEYLAB_BIO_PHYLOGENETIC_TREE_HH__
