#!/usr/bin/env python

# Copyright (c) 2007
# Colin Dewey (University of Wisconsin-Madison)
# cdewey@biostat.wisc.edu
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


import sys
import os 
import re
from Bio import Fasta

from SeqUtil import isTupleOrList

def makeEncodingTable(chars):
    return dict(zip(chars, range(len(chars))))

def invertTable(table):
    return dict(zip(table.values(), table.keys()))

def makeCodonTable(base1, base2, base3, aas):
    return dict(zip(map(''.join, zip(base1, base2, base3)), aas))

def makeMatrix(dims, value=None):
    if len(dims) == 1:
        return [value] * dims[0]
    else:
        return [makeMatrix(dims[1:], value) for i in xrange(dims[0])]

def uniqueElts(seq):
    return dict(zip(seq, [None] * len(seq))).keys()

def quoteName(name):
    if name is None:
        return "NULL"
    else:
        return '"%s"' % name

def makeNameArray(names):
    return ",\n".join(map(quoteName, names))

def quoteChar(c):
    return "'%s'" % c

def makeMatrixString(matrix):
    if isTupleOrList(matrix[0]):
        return "{\n%s}" % ",\n".join(map(makeMatrixString, matrix))
    else:
        return "{%s}" % ",".join(map(quoteChar, matrix))

def makeTableName(id):
    if id is None:
        return "NULL"
    else:
        return "&TRANSLATION_TABLE%d[0][0][0]" % id

def makeTableArray(ids):
    idDict = dict(zip(ids, ids))
    idList = [idDict.get(i) for i in xrange(0, max(ids) + 1)]
    return ",\n".join(map(makeTableName, idList))

if len(sys.argv) != 3:
    print "Usage: %s tablefile.prt testdir" \
          % os.path.basename(sys.argv[0])
    sys.exit(1)

tableText = open(sys.argv[1]).read()
testdir = sys.argv[2]

tablePat = re.compile(r'\{\s*' +
                      r'name\s*"(?P<name>[^"]*)"\s*,\s*' +
                      r'(?:name\s*"(?P<name2>[^"]*)"\s*,\s*)?' +
                      r'id\s*(?P<id>\d+)\s*,\s*' +
                      r'ncbieaa\s*"(?P<aas>[^"]*)"\s*,\s*' +
                      r'sncbieaa\s*"(?P<starts>[^"]*)"\s*' +
                      r'-- Base1\s*(?P<base1>\S*)\s*' +
                      r'-- Base2\s*(?P<base2>\S*)\s*' +
                      r'-- Base3\s*(?P<base3>\S*)\s*' +
                      r'\}')

dna = "ACGTUNMRWSYKVHDB"
dnaEncoding = makeEncodingTable(dna)
dnaAmbiguities = {
    "A": "A",
    "C": "C",
    "G": "G",
    "T": "T",
    "U": "T",
    "M": "AC",
    "R": "AG",
    "W": "AT",
    "S": "CG",
    "Y": "CT",
    "K": "GT",
    "V": "ACG",
    "H": "ACT",
    "D": "AGT",
    "B": "CGT",
    "N": "GATC",
    }

ambiguousToAas = {
    "A": "A",
    "B": "DN",
    "C": "C",
    "D": "D",
    "E": "E",
    "F": "F",
    "G": "G",
    "H": "H",
    "I": "I",
    "K": "K",
    "L": "L",
    "M": "M",
    "N": "N",
    "P": "P",
    "Q": "Q",
    "R": "R",
    "S": "S",
    "T": "T",
    "V": "V",
    "W": "W",
    "X": "ACDEFGHIKLMNPQRSTVWY",
    "Y": "Y",
    "Z": "EQ",
    "*": "*",
    }

aasToAmbiguous = invertTable(ambiguousToAas)

ambiguousCodonTable = {}
for b1 in dna:
    for b2 in dna:
        for b3 in dna:
            unambiguousCodons = [c1 + c2 + c3
                                 for c1 in dnaAmbiguities[b1]
                                 for c2 in dnaAmbiguities[b2]
                                 for c3 in dnaAmbiguities[b3]]
            ambiguousCodonTable[b1 + b2 + b3] = unambiguousCodons

ambiguousCodons = ambiguousCodonTable.keys()
ambiguousCodons.sort()
codonFile = open(os.path.join(testdir, "codons.fa"), 'w')
codonRec = Fasta.Record()
codonRec.title = ""
codonRec.sequence = ''.join(ambiguousCodons)
codonFile.write(str(codonRec) + '\n')
codonFile.close()

names1 = {}
names2 = {}
matrices = {}

for match in tablePat.finditer(tableText):
    codonTable = makeCodonTable(match.group("base1"),
                                match.group("base2"),
                                match.group("base3"),
                                match.group("aas"))
    codonMatrix = makeMatrix((16, 16, 16), '-')
    codonMap = {}
    for aCodon, codons in ambiguousCodonTable.items():
        aas = uniqueElts([codonTable[c] for c in ambiguousCodonTable[aCodon]])
        aas.sort()
        aas = ''.join(aas)
        if aas in aasToAmbiguous:
            aa = aasToAmbiguous[aas]
        else:
            aa = 'X'
        indices = [dnaEncoding[char] for char in aCodon]
        codonMatrix[indices[0]][indices[1]][indices[2]] = aa
        codonMap[aCodon] = aa

    id = int(match.group("id"))
    matrices[id] = codonMatrix
    names1[id] = match.group("name").replace('\n', '')
    names2[id] = match.group("name2")

    proteinFile = open(os.path.join(testdir, "aas%d.fa" % id), 'w')
    proteinRec = Fasta.Record()
    proteinRec.title = ""
    proteinRec.sequence = ''.join([codonMap[ac] for ac in ambiguousCodons])
    proteinFile.write(str(proteinRec) + '\n')
    proteinFile.close()

ids = names1.keys()
ids.sort()

for id in ids:
    print "const char TRANSLATION_TABLE%d[16][16][16] = %s;" % \
          (id, makeMatrixString(matrices[id]))
    print

for id in ids:
    name1 = names1[id]
    if name1 is None:
        name1 = ""
    name2 = names2[id]
    if name2 is None:
        name2 = ""
    print "static const Table table%d(&TRANSLATION_TABLE%d[0][0][0], %d, %s, %s);" % \
          (id, id, id, '"%s"' % name1, '"%s"' % name2)
