/* Copyright (c) 2010
   Colin Dewey (University of Wisconsin-Madison)
   cdewey@biostat.wisc.edu
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "config.h"

#include "deweylab/bio/formats/fastq.hh"
#include "deweylab/util/options.hh"
#include "deweylab/bio/alphabet/AmbiguousNucleotide.hh"

#include <iostream>
#include <algorithm>

using namespace deweylab;
using namespace deweylab::bio::formats;

class ReverseComplementer {
public:
    ReverseComplementer(const bio::alphabet::Nucleotide* alphabet)
        : alphabet_(alphabet) {
    }

    fastq::Record operator()(const fastq::Record& rec) {
        fastq::Record rc_rec(rec);
        alphabet_->reverseComplementInPlace(rc_rec.sequence);
        reverse(rc_rec.qualities.begin(), rc_rec.qualities.end());
        return rc_rec;
    }
    
private:
    const bio::alphabet::Nucleotide* alphabet_;
};

int main(int argc, const char* argv[]) {
    std::string encoding_name = "sanger";
    const bio::alphabet::AmbiguousNucleotide* alphabet
        = &bio::alphabet::AMBIGUOUS_DNA;
    
    deweylab::util::options::Parser parser("< fastq input");
    parser.setDescription("Reverse-complement FASTQ-formatted sequences");
    parser.addStoreOpt('e', "encoding",
                       "Quality score encoding scheme",
                       encoding_name);
	parser.addStoreConstOpt('r', "rna",
                            "Treat sequence as RNA instead of DNA",
                            alphabet, &bio::alphabet::AMBIGUOUS_RNA);
	parser.parse(argv, argv + argc);
    
	try {
        const fastq::QualityScoreEncoding* encoding =
            fastq::get_quality_score_encoding(encoding_name);

        fastq::InputStream in_stream(std::cin, encoding);
        fastq::OutputStream out_stream(std::cout, encoding);
        fastq::InputStreamIterator in_begin(in_stream), in_end;
        fastq::OutputStreamIterator out_begin(out_stream);
        ReverseComplementer reverse_complementer(alphabet);
        std::transform(in_begin, in_end, out_begin, reverse_complementer);
    } catch (const std::runtime_error& e) {
		std::cerr << "Error: " << e.what() << '\n';
		return EXIT_FAILURE;
	}

	return EXIT_SUCCESS;
}
