/* Copyright (c) 2010
   Colin Dewey (University of Wisconsin-Madison)
   cdewey@biostat.wisc.edu
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "config.h"

#include "deweylab/bio/formats/fastq.hh"
#include "deweylab/util/options.hh"

#include <iostream>

using namespace deweylab::bio::formats;

class Trimmer {
public:
    Trimmer(size_t max_read_length,
            size_t five_prime_trim,
            size_t three_prime_trim,
            size_t min_quality) 
        : max_read_length_(max_read_length),
          five_prime_trim_(five_prime_trim),
          three_prime_trim_(three_prime_trim),
          min_quality_(min_quality) {
    }

    fastq::Record operator()(const fastq::Record& rec) {
        size_t start = (five_prime_trim_ <= rec.sequence.size() ?
                        five_prime_trim_ : 0);
        size_t end = (three_prime_trim_ <= rec.sequence.size() ?
                      rec.sequence.size() - three_prime_trim_ : 0);
        if (min_quality_ > 0) {
            char min_quality_char =
                rec.qualities_encoding->encode_phred(min_quality_);
            int best_start = -1;
            int best_length = 0;
            int best_total = 0;
            int curr_start = -1;
            int curr_total = 0;
            for (size_t i = start; i <= end; ++i) {
                if (i < end && rec.qualities[i] >= min_quality_char) {
                    if (curr_start == -1) {
                        curr_start = i;
                    }
                    curr_total += rec.qualities[i];
                } else {
                    if (curr_start != -1) {
                        int curr_length = i - curr_start;
                        if (curr_length > best_length ||
                            (curr_length == best_length &&
                             curr_total > best_total)) {
                            best_start = curr_start;
                            best_length = curr_length;
                            best_total = curr_total;
                        }
                    }
                    curr_start = -1;
                    curr_total = 0;
                }
            }

            if (best_start != -1) {
                start = best_start;
                end = best_start + best_length;
            } else {
                start = 0;
                end = 0;
            }
        }

        fastq::Record trimmed(rec);
        size_t length = end - start;
        if (max_read_length_ > 0) {
            length = std::min(length, max_read_length_);
        }
        trimmed.sequence = trimmed.sequence.substr(start, length);
        trimmed.qualities = trimmed.qualities.substr(start, length);
        return trimmed;
    }
    
private:
    size_t max_read_length_;
    size_t five_prime_trim_;
    size_t three_prime_trim_;
    size_t min_quality_;
};

int main(int argc, const char* argv[]) {
    size_t max_read_length = 0;
    size_t five_prime_trim = 0;
    size_t three_prime_trim = 0;
    size_t min_quality = 0;
    std::string encoding_name = "sanger";

    deweylab::util::options::Parser parser("< fastq input");
    parser.setDescription("Trim FASTQ-formatted sequences");
    parser.addStoreOpt('e', "encoding",
                       "Quality score encoding scheme",
                       encoding_name);
    parser.addStoreOpt('l', "length",
                       "Trim additional bases from the 3' end so that the "
                       "resulting reads have length no greater than the "
                       "specified value.  If this is set to zero, then this "
                       "behavior is disabled.",
                       max_read_length);
    parser.addStoreOpt('3', "3-prime-trim",
                       "Number of bases to trim from the 3' end",
                       three_prime_trim);
    parser.addStoreOpt('5', "5-prime-trim",
                       "Number of bases to trim from the 5' end",
                       five_prime_trim);
    parser.addStoreOpt('q', "min-quality",
                       "Trim to keep the longest interval of the sequence "
                       "that only contains positions with phred quality "
                       "greater than the specified value.  In the case of ties, "
                       "the interval with the larger mean quality value is chosen, "
                       "or if all intervals have the same mean quality value, "
                       "the 5'-most interval is chosen.",
                       min_quality);
	parser.parse(argv, argv + argc);
    
	try {
        const fastq::QualityScoreEncoding* encoding =
            fastq::get_quality_score_encoding(encoding_name);
        fastq::InputStream in_stream(std::cin, encoding);
        fastq::OutputStream out_stream(std::cout, encoding);
        fastq::InputStreamIterator in_begin(in_stream), in_end;
        fastq::OutputStreamIterator out_begin(out_stream);
        Trimmer trimmer(max_read_length,
                        five_prime_trim,
                        three_prime_trim,
                        min_quality);
        std::transform(in_begin, in_end, out_begin, trimmer);
    } catch (const std::runtime_error& e) {
		std::cerr << "Error: " << e.what() << '\n';
		return EXIT_FAILURE;
	}

	return EXIT_SUCCESS;
}
