#!/usr/bin/env python

# Copyright (c) 2007
# Colin Dewey (University of Wisconsin-Madison)
# cdewey@biostat.wisc.edu
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import sys
from collections import defaultdict
from deweylab.bio.formats import gff

if len(sys.argv) != 2:
    print >>sys.stderr, "Usage: %s sourceName < ucsc_input > gff_output" \
          % sys.argv[0]
    sys.exit(1)

source = sys.argv[1]

# Dictionary to keep track of transcript numbers
name_counts = defaultdict(int)

writer = gff.Writer(sys.stdout, version="2.2")

# Possible characters for the strand field
strands = ('+', '-')

for line in sys.stdin:
    # Skip header line if downloaded from Table Browser
    if line.startswith("#"): continue

    fields = line[:-1].split('\t')

    # Determine field structure based on which field has the strand
    if not fields[2] in strands:
        if fields[3] in strands:
            # Shift over one field due to 'bin' first field
            fields = fields[1:]
        else:
            print >>sys.stderr, "Unrecognized UCSC gene format"
            sys.exit(1)
    
    (name, chrom, strand, txStart, txEnd, cdsStart, cdsEnd, \
     exonCount, exonStarts, exonEnds) = fields[0:10]

    if len(fields) == 15:
        gene_id = fields[11]
    else:
        gene_id = name

    (txStart, txEnd, cdsStart, cdsEnd, exonCount) = \
              map(long, (txStart, txEnd, cdsStart, cdsEnd, exonCount))

    name_counts[name] += 1
    transcript_id = "%s.%d" % (name, name_counts[name])

    rec = gff.Record(chrom,
                     source,
                     None,
                     None,
                     None,
                     score=None,
                     strand=strand,
                     phase=None,
                     attributes={"gene_id": [gene_id],
                                 "transcript_id": [transcript_id]})

    features = []

    if cdsStart != cdsEnd:
        rec.type = "start_codon"
        if strand == '+':
            rec.start = cdsStart + 1
            rec.end = cdsStart + 3
        else:
            rec.start = cdsEnd - 2
            rec.end = cdsEnd
        features.append(rec.copy())    

    # Generate gff lines for exons
    exonStarts = map(long, exonStarts.split(',')[:-1])
    exonEnds = map(long, exonEnds.split(',')[:-1])
    if strand == '-':
        exonStarts.reverse()
        exonEnds.reverse()
        
    phase = 0
    for (exonNum, exonStart, exonEnd) in zip(xrange(exonCount),
                                             exonStarts,
                                             exonEnds):
        if exonNum == 0:
            rec.attributes["exontype"] = ["initial"]
        elif exonNum == (exonCount - 1):
            rec.attributes["exontype"] = ["terminal"]
        else:
            rec.attributes["exontype"] = ["internal"]

        rec.type = "exon"
        rec.start = exonStart + 1
        rec.end = exonEnd
        rec.phase = None
        features.append(rec.copy())

        exonCodingStart = max(exonStart, cdsStart)
        exonCodingEnd = min(exonEnd, cdsEnd)
        if (exonCodingStart < exonCodingEnd):
            rec.type = "CDS"
            rec.start = exonCodingStart + 1
            rec.end = exonCodingEnd
            rec.phase = (3 - phase) % 3
            phase = (phase + exonCodingEnd - exonCodingStart) % 3
            features.append(rec.copy())                    

    if cdsStart != cdsEnd:
        rec.type = "stop_codon"
        rec.phase = None
        del rec.attributes["exontype"]
        if strand == '+':
            rec.start = cdsEnd - 2
            rec.end = cdsEnd
        else:
            rec.start = cdsStart + 1
            rec.end = cdsStart + 3
        features.append(rec.copy())        

    # Write GFF lines for this gene
    writer.write_recs(features)
