/* Copyright (c) 2007
   Colin Dewey (University of Wisconsin-Madison)
   cdewey@biostat.wisc.edu
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <iostream>
#include <map>
#include <vector>

#include "deweylab/bio/formats/agp/Record.hh"
#include "deweylab/bio/formats/agp/InputStream.hh"
#include "deweylab/bio/formats/fasta/DB.hh"
#include "deweylab/bio/formats/fasta/OutputStream.hh"
#include "deweylab/util/stl.hh"
#include "deweylab/util/options.hh"
#include "deweylab/filesystem/InputFileStream.hh"
using namespace deweylab;
using namespace bio::formats;

int main(int argc, const char* argv[]) {
	// Increase speed of input/output to standard streams
	std::ios::sync_with_stdio(false);

	// Initialize options to defaults
	std::string agpFilename;
	
	util::options::Parser parser("< fastaInput > fastaOutput",
								 "Assemble sequences from input into larger "
								 "sequences as specified by an AGP file");
	parser.addStoreArg("apgFile",
					   "AGP file specifying how the sequences in the input "
					   "are assembled into sequences in the output",
					   agpFilename);
	parser.parse(argv, argv + argc);

	try {
		fasta::DB seq_db(std::cin);
			
		// Read AGP records into map
		typedef std::vector<agp::Record> AGPRecList;
		typedef std::map<std::string, AGPRecList> AGPRecMap;
		AGPRecMap chromRecs;
		filesystem::InputFileStream agpFile(agpFilename);
		agp::InputStream agpStream(agpFile);
		agp::Record rec;
		while (agpStream >> rec) {
			chromRecs[rec.getChrom()].push_back(rec);
		}
	
		// Construct each chromosome separately and insert into output
		bio::formats::fasta::OutputStream fastaOutStream(std::cout);
		AGPRecMap::iterator it;
		for (it = chromRecs.begin(); it != chromRecs.end(); ++it) {
			const std::string& chrom = it->first;
			AGPRecList& recs = it->second;
	
			// Order the AGP records by contig start coordinate
			std::sort(recs.begin(), recs.end());
	
			// Determine the chrom length and make a new sequence
			std::string chromSeq(recs.back().getTargetEnd(), 'N');
			AGPRecList::iterator recIt;
			for (recIt = recs.begin(); recIt != recs.end(); ++recIt) {
				if (recIt->isGap()) { continue; }
	
				std::string seq = seq_db.getIntervalSubseq(recIt->getSourceInterval());
 				chromSeq.replace(recIt->getTargetStart(), seq.size(), seq);
			}
	
			fastaOutStream << fasta::Record(chrom, "", chromSeq);
		}
	} catch (const std::runtime_error& e) {
		std::cerr << "Error: " << e.what() << '\n';
		return EXIT_FAILURE;
	}		
	
	return EXIT_SUCCESS;
}
