/* Copyright (c) 2007
   Colin Dewey (University of Wisconsin-Madison)
   cdewey@biostat.wisc.edu
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <iostream>
#include <iomanip>

#include "deweylab/bio/formats/fasta.hh"
#include "deweylab/bio/alphabet/AmbiguousNucleotide.hh"
#include "deweylab/bio/alphabet/AmbiguousAminoAcid.hh"
#include "deweylab/bio/alphabet/AlphabetCounter.hh"
#include "deweylab/util/string.hh"
#include "deweylab/util/options.hh"
using namespace deweylab;
using namespace bio::alphabet;

void output_counts(std::ostream& stream,
				   const AlphabetCounter& counter,
				   bool show_freqs) {
	for (size_t i = 0; i < counter.getAlphabet().getSize(); ++i) {
		std::cout << '\t';
		if (show_freqs) {
			std::cout << std::setprecision(2) << counter.getEncodedFraction(i);
		} else {
			std::cout << counter.getEncodedCount(i);
		}
	}
}

int main(int argc, const char* argv[]) {
	// Increase speed of input/output to standard streams
	std::ios::sync_with_stdio(false);

	// Initialize options to defaults
	const bio::alphabet::Alphabet* alphabet = &bio::alphabet::AMBIGUOUS_DNA;
	bool show_freqs = false;
	bool show_header = false;
	bool show_total = false;
	bool show_names = false;
	bool total_only = false;

	// Set up options
	util::options::Parser parser("< fastaInput",
								 "Calculates composition counts of FASTA sequences");
	parser.addStoreConstOpt('d', "dna", "Input sequences are DNA (default)",
							alphabet,
							static_cast<const Alphabet*>(&AMBIGUOUS_DNA));
	parser.addStoreConstOpt('r', "rna", "Input sequences are RNA",
							alphabet,
							static_cast<const Alphabet*>(&AMBIGUOUS_RNA));
	parser.addStoreConstOpt('p', "protein", "Input sequences are protein",
							alphabet,
							static_cast<const Alphabet*>(&AMBIGUOUS_PROTEIN));
	parser.addStoreTrueOpt('f', "freqs", "Show frequencies instead of counts",
							show_freqs);
	parser.addStoreTrueOpt('h', "header", "Display header line",
							show_header);
	parser.addStoreTrueOpt('t', "total", "Display totals line",
							show_total);
	parser.addStoreTrueOpt('n', "names", "Display full residue names",
							show_names);
	parser.addStoreTrueOpt(0, "total-only", "Only display totals line",
						   total_only);
	parser.parse(argv, argv + argc);

	try {
		// Initialize counts
		AlphabetCounter seq_counter(*alphabet);
		AlphabetCounter totals_counter(*alphabet);
	
		// Optionally output header
		if (show_header) {
			std::cout << "name" << '\t' << "length";
			for (size_t i = 0; i < alphabet->getSize(); ++i) {
				std::cout << '\t';
				char residue = alphabet->decode(i);
				if (show_names) {
					std::cout << alphabet->getName(residue);
				} else {
					std::cout << residue;
				}
			}
			std::cout << '\n';
		}

		// Construct FASTA stream for fast reading
		bio::formats::fasta::InputStream fasta_stream(std::cin);
	
		// Step through records
		bio::formats::fasta::Record rec;
		while (fasta_stream >> rec) {
			std::string::const_iterator begin = rec.sequence.begin();
			std::string::const_iterator end = rec.sequence.end();
			if (alphabet == &bio::alphabet::AMBIGUOUS_PROTEIN and
				not (begin == end) and *(end - 1) == '*') {
				--end;
			}
			if (not alphabet->isOn(begin, end)) {
				throw std::runtime_error("Sequence is not of specified alphabet: "
										 + rec.sequence);
			}

			seq_counter.reset();
			seq_counter.count(begin, end);
			totals_counter.addCounts(seq_counter);

			if (not total_only) {
				// Output record counts
				std::cout << rec.id << '\t' << seq_counter.getTotal();
				output_counts(std::cout, seq_counter, show_freqs);
				std::cout << '\n';
			}
		}

		// Optionally show total counts
		if (total_only or show_total) {
			std::cout << "totals" << '\t' << totals_counter.getTotal();
			output_counts(std::cout, totals_counter, show_freqs);
			std::cout << '\n';
		}

	} catch (const std::runtime_error& e) {
		std::cerr << "Error: " << e.what() << '\n';
		return EXIT_FAILURE;
	}

	return EXIT_SUCCESS;
}
