/* Copyright (c) 2007
   Colin Dewey (University of Wisconsin-Madison)
   cdewey@biostat.wisc.edu
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <iostream>

#include "deweylab/bio/formats/fasta.hh"
#include "deweylab/bio/alphabet/AmbiguousNucleotide.hh"
#include "deweylab/bio/alphabet/AmbiguousAminoAcid.hh"
#include "deweylab/bio/alphabet/AlphabetCounter.hh"
#include "deweylab/util/string.hh"
#include "deweylab/util/options.hh"
using namespace deweylab;

class CharFilter {
public:
	CharFilter(char c, bool greater_than, double fraction)
		: c(c), greater_than(greater_than), fraction(fraction) {
	}

	bool passes(const bio::alphabet::AlphabetCounter& counter) const {
		if (greater_than) {
			return counter.getFraction(c) >= fraction;
		} else {
			return counter.getFraction(c) <= fraction;
		}
	}
	
private:
	char c;
	bool greater_than;
	double fraction;
};

class CompFilter {
public:
	CompFilter(const std::string& filter_string,
			   const bio::alphabet::Alphabet& alphabet)
		: alphabet(alphabet), counter(alphabet), filters() {
		parseFilterString(filter_string); 
	}

	bool passes(const std::string& s) {
		std::string::const_iterator begin = s.begin();
		std::string::const_iterator end = s.end();
		if (&alphabet == &bio::alphabet::AMBIGUOUS_PROTEIN and
			not s.empty() and s[s.size() - 1] == '*') {
			--end;
		}
		if (not alphabet.isOn(begin, end)) {
			throw std::runtime_error("Sequence is not of specified alphabet: "
									 + s);
		}
		counter.reset();
		counter.count(begin, end);
		for (size_t i = 0; i < filters.size(); ++i) {
			if (not filters[i].passes(counter)) {
				return false;
			}
		}
		return true;
	}
	
private:
	const bio::alphabet::Alphabet& alphabet;
	bio::alphabet::AlphabetCounter counter;
	std::vector<CharFilter> filters;

	void parseFilterString(const std::string& filter_string) {
		std::vector<std::string> tokens;
		util::string::split(filter_string, std::back_inserter(tokens));
		if (tokens.size() % 3 != 0) {
			throw std::runtime_error("Invalid filter format");
		}
		for (size_t i = 0; i < tokens.size(); i += 3) {
			parseFilterLine(tokens[i], tokens[i + 1], tokens[i + 2]);
		}
	}

	void parseFilterLine(const std::string& c,
						 const std::string& op,
						 const std::string& fraction) {
		if (c.size() != 1 or not alphabet.isMember(c[0])) {
			throw std::runtime_error("Invalid residue:" + c);
		}

		if (op != ">" and op != "<") {
			throw std::runtime_error("Invalid operator:" + op);
		}

		util::string::Converter<double> toDouble;

		filters.push_back(CharFilter(c[0], op == ">", toDouble(fraction)));
	}
};

int main(int argc, const char* argv[]) {
	// Increase speed of input/output to standard streams
	std::ios::sync_with_stdio(false);

	// Initialize options to defaults
	bool protein = false;
	std::string filter_string;
	std::string filter_filename;

	// Set up options
	util::options::Parser parser("< fastaInput",
								 "Filter FASTA-formatted sequences based on residue composition");
	parser.addStoreTrueOpt('p', "protein", "Sequences are proteins", protein);
	parser.addStoreOpt('f', "filter", "Filter string", filter_string);
	parser.addStoreOpt(0, "filter-file", "File containing filter string",
					   filter_filename);
	parser.parse(argv, argv + argc);

	try {
		const bio::alphabet::Alphabet* alphabet;
		if (protein) {
			alphabet = &bio::alphabet::AMBIGUOUS_PROTEIN;
		} else {
			alphabet = &bio::alphabet::AMBIGUOUS_DNA;
		}
		
		CompFilter filter(filter_string, *alphabet);
		
		bio::formats::fasta::InputStream in_stream(std::cin);
		bio::formats::fasta::OutputStream out_stream(std::cout);
		
		// Step through records
		bio::formats::fasta::Record rec;
		while (in_stream >> rec) {
			if (filter.passes(rec.sequence)) {
				out_stream << rec;
			}
		}
	} catch (const std::runtime_error& e) {
		std::cerr << "Error: " << e.what() << '\n';
		return EXIT_FAILURE;
	}

	return EXIT_SUCCESS;
}
