/* Copyright (c) 2007
   Colin Dewey (University of Wisconsin-Madison)
   cdewey@biostat.wisc.edu
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <string>
#include <iostream>  // for I/O
#include <stdexcept> // for runtime_error
#include <set>

#include "boost/shared_ptr.hpp"

#include "deweylab/bio/formats/fasta.hh"
#include "deweylab/util/string.hh"
#include "deweylab/util/options.hh"
#include "deweylab/filesystem.hh"
using namespace deweylab;
using namespace filesystem;
using namespace bio::formats;

typedef std::vector<fasta::Record> RecList;
typedef boost::shared_ptr<RecList> RecListPtr;
typedef RecList::iterator RecListIter;

struct FastaRecordSet {
    RecListPtr recs;

    FastaRecordSet() : recs(new RecList()) {}
};

typedef std::vector<FastaRecordSet> RecSetList;
typedef RecSetList::iterator RecSetListIter;

struct RecordBinWriter {
    const Path __dir;
    const std::string __ext;
    const std::string __prefix;
    bool __quiet;
    bool __use_title;
    size_t __num;
    
    RecordBinWriter(const Path& dir,
                    const std::string& ext,
                    const std::string& prefix,
                    bool quiet,
                    bool use_title)
        : __dir(dir), __ext(ext), __prefix(prefix),
          __quiet(quiet), __use_title(use_title),
          __num(0) {
    }

    void operator()(const fasta::Record& rec);
    void operator()(const FastaRecordSet& set);
};

void RecordBinWriter::operator()(const fasta::Record& rec) {
    std::string rec_name;
    if (__use_title) {
        // Construct name of record file from id
        rec_name = util::string::validPosix(rec.id);
        if (rec_name == "") {
            throw std::runtime_error("Could not write file for record with "
                                     "description: " + rec.id);
        }
    } else {
        rec_name = __prefix + util::string::toString(__num);
        ++__num;
    }
    
	// Open record file and write record
	Path path(__dir / (rec_name + __ext));
	OutputFileStream writeFile(path);
	fasta::OutputStream fastaOutStream(writeFile);
	fastaOutStream << rec;
    if (not __quiet) {
        std::cerr << path.toString() << '\n';
    }
}

void RecordBinWriter::operator()(const FastaRecordSet& set) {
    std::string rec_name = __prefix + util::string::toString(__num);
    ++__num;

	// Open record file and write record
	Path path(__dir / (rec_name + __ext));
    
	OutputFileStream writeFile(path);
	fasta::OutputStream fastaOutStream(writeFile);
    fasta::OutputStreamIterator out_iter(fastaOutStream);
    std::copy(set.recs->begin(), set.recs->end(), out_iter);
    if (not __quiet) {
        std::cout << path.toString() << '\n';
    }
}

struct FastaRecordLargerSize {
    bool operator()(const fasta::Record& r1, const fasta::Record& r2) const {
        return r1.sequence.size() > r2.sequence.size();
    }
};

struct IndexSizePair {
    size_t i;
    size_t size;
    IndexSizePair(size_t i = 0, size_t size = 0) : i(i), size(size) {}
};

struct CapacityComparer {
    bool operator()(const IndexSizePair& p1, const IndexSizePair& p2) const {
        return p1.size > p2.size;
    }
};

void pack_records_by_num_files(RecList& recs,
                               RecSetList& sets,
                               size_t num_files) {
    std::vector<IndexSizePair> pairs;
    for (size_t i = 0; i < num_files; ++i) {
        sets.push_back(FastaRecordSet());
        pairs.push_back(IndexSizePair(i, 0));
    }

    for (RecListIter it = recs.begin(); it != recs.end(); ++it) {
        std::pop_heap(pairs.begin(), pairs.end(), CapacityComparer());
        sets[pairs.back().i].recs->push_back(*it);
        pairs.back().size += it->sequence.size();
        std::push_heap(pairs.begin(), pairs.end(), CapacityComparer());
    }
}

void pack_records_by_max_size(RecList& recs,
                              RecSetList& sets,
                              size_t max_size) {
    typedef std::multiset<IndexSizePair, CapacityComparer> PairSet;
    typedef PairSet::iterator PairSetIter;

    PairSet pairs;

    for (RecListIter it = recs.begin(); it != recs.end(); ++it) {
        size_t seq_size = it->sequence.size();
        
        assert(seq_size <= max_size);
        
        // Find largest set with enough capacity for this record
        PairSetIter pair_it =
            pairs.lower_bound(IndexSizePair(0, max_size - seq_size));

        IndexSizePair pair;

        // If no such set exists, create a new one
        if (pair_it == pairs.end()) {
            pair.i = sets.size();
            sets.push_back(FastaRecordSet());
        } else {
            pair = *pair_it;
            pairs.erase(pair_it);
        }

        pair.size += seq_size;
        pairs.insert(pair);
        sets[pair.i].recs->push_back(*it);
    }
}

int main(int argc, const char* argv[]) {
	// Increase speed of input/output to standard streams
	std::ios::sync_with_stdio(false);

	std::string outDirString = ".";
	std::string extension = ".fasta";
    std::string prefix = "bin";
    bool quiet = false;
    bool numbered_files = false;
    size_t num_files = 0;
    size_t max_size = 0;

	// Parse command line
	util::options::Parser parser("< fastaInput",
								 "Split single FASTA stream into multiple "
                                 "files.  Default behavior is to split into "
                                 "one file per record.  With the -n or -s "
                                 "options, records may be binned into "
                                 "multi-record files.");
	parser.addStoreOpt('d', "out-dir",
					   "Directory in to which to output the FASTA files",
					   outDirString, "DIR");
	parser.addStoreOpt('n', "num-files",
					   "Split input into exactly N files",
					   num_files, "N");
	parser.addStoreOpt('s', "max-size",
					   "Split input into files such that the maximum size "
                       "(sum of sequence lengths) of any file is less than S",
					   max_size, "S");
	parser.addStoreOpt('e', "ext", "Extension for FASTA filenames", extension);
	parser.addStoreOpt('p', "prefix", "Prefix for FASTA filenames", prefix);
	parser.addStoreTrueOpt(0, "numbered-files",
                           "Create numbered output files. "
                           "This is always on if -n or -s is specified. "
                           "Otherwise the default is to name files by the "
                           "title of the record contained within it.",
                           numbered_files);
	parser.addStoreTrueOpt('q', "quiet", "Do not output names of files created", quiet);
	parser.parse(argv, argv + argc);

	try {
		Path out_dir(outDirString);
		if (not out_dir.exists() or not out_dir.isDirectory()) {
			throw std::runtime_error("Invalid output directory: "
									 + out_dir.toString());
		}

        fasta::InputStream in_stream(std::cin);
        fasta::InputStreamIterator begin(in_stream), end;
        RecordBinWriter writer(out_dir, extension, prefix, quiet,
                               not numbered_files);
        
        if (num_files > 0 or max_size > 0) {
            RecList recs;
            RecSetList sets;
            
            std::copy(begin, end, std::back_inserter(recs));
            std::sort(recs.begin(), recs.end(), FastaRecordLargerSize());
            if (num_files > 0) {
                pack_records_by_num_files(recs, sets, num_files);
            } else {
                pack_records_by_max_size(recs, sets, max_size);
            }
            std::for_each(sets.begin(), sets.end(), writer);
        } else {
            std::for_each(begin, end, writer);
        }

	} catch (const std::runtime_error& e) {
		std::cerr << "Error: " << e.what() << '\n';
		return EXIT_FAILURE;
	}

	return EXIT_SUCCESS;
}
