/* Copyright (c) 2010
   Colin Dewey (University of Wisconsin-Madison)
   cdewey@biostat.wisc.edu
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "config.h"

#include "deweylab/util/options.hh"
#include "deweylab/filesystem.hh"

#include "deweylab/bio/formats/fasta.hh"
#include "deweylab/bio/formats/qual.hh"
#include "deweylab/bio/formats/fastq.hh"

#include <vector>
#include <string>

using deweylab::bio::formats::fastq::QualityScoreEncoding;

class FastaToFastqConverter {
public:
    FastaToFastqConverter(int quality_score,
                          const QualityScoreEncoding* encoding)
        : quality_char_(encoding->encode_phred(quality_score)),
          encoding_(encoding) {
    }

    void run();

private:
    char quality_char_;
    const QualityScoreEncoding* encoding_;
};

void FastaToFastqConverter::run() {
    using namespace deweylab::filesystem;
    using namespace deweylab::bio::formats;
    fasta::InputStream fasta_in(std::cin);
    fastq::OutputStream fastq_out(std::cout, encoding_);

    fasta::Record fasta_rec;
    fastq::Record fastq_rec;

    fastq_rec.qualities_encoding = encoding_;

    while (fasta_in >> fasta_rec) {
        fastq_rec.id = fasta_rec.id;
        fastq_rec.description = fasta_rec.description;
        fastq_rec.sequence = fasta_rec.sequence;
        fastq_rec.qualities = std::string(fasta_rec.sequence.size(),
                                          quality_char_);
        // Output the FASTQ record
        fastq_out << fastq_rec;
    }
}

class FastaQualToFastqConverter {
public:
    FastaQualToFastqConverter(const std::string& qual_filename,
                              const QualityScoreEncoding* encoding)
        : qual_filename_(qual_filename),
          encoding_(encoding) {
    }

    void run();

private:
    std::string qual_filename_;
    const QualityScoreEncoding* encoding_;
};

void FastaQualToFastqConverter::run() {
    using namespace deweylab::filesystem;
    using namespace deweylab::bio::formats;
    fasta::InputStream fasta_in(std::cin);
    InputFileStream qual_file(qual_filename_);
    qual::InputStream qual_in(qual_file);
    fastq::OutputStream fastq_out(std::cout, encoding_);

    fasta::Record fasta_rec;
    qual::Record qual_rec;
    fastq::Record fastq_rec;

    fastq_rec.qualities_encoding = encoding_;
    
    while (true) {
        // Read the next records from both streams
        fasta_in >> fasta_rec;
        qual_in >> qual_rec;

        // Stop if we reach the end of either stream
        if (!fasta_in || !qual_in) break;

        // Check that the records match
        if (fasta_rec.id != qual_rec.id) {
            throw std::runtime_error("Titles of FASTA and QUAL records do not "
                                     "match:\n" +
                                     fasta_rec.id + "\n" + qual_rec.id);
        }

        // Construct FASTQ record by merging the two records
        fastq_rec.id = fasta_rec.id;
        fastq_rec.description = fasta_rec.description;
        fastq_rec.sequence = fasta_rec.sequence;
        fastq_rec.set_phred_qualities(qual_rec.qualities);

        // Output the FASTQ record
        fastq_out << fastq_rec;
    }
}

int main(int argc, const char* argv[]) {
    std::string qual_filename;
    std::string output_encoding_name = "sanger";
    int quality_score = 40;

    deweylab::util::options::Parser parser;
    parser.addStoreOpt('q', "quals-file",
                       "File containing quality scores in QUAL format",
                       qual_filename);
    parser.addStoreOpt('s', "quality-score",
                       "Default phred quality score if no qualities file is "
                       "provided",
                       quality_score);
    parser.addStoreOpt('e', "encoding",
                       "Quality score encoding scheme",
                       output_encoding_name);
	parser.parse(argv, argv + argc);
    
	try {
        using namespace deweylab::bio::formats::fastq;
        const QualityScoreEncoding* output_encoding = get_quality_score_encoding(output_encoding_name);
        if (not qual_filename.empty()) {
            FastaQualToFastqConverter converter(qual_filename, output_encoding);
            converter.run();
        } else {
            FastaToFastqConverter converter(quality_score, output_encoding);
            converter.run();
        }
    } catch (const std::runtime_error& e) {
		std::cerr << "Error: " << e.what() << '\n';
		return EXIT_FAILURE;
	}

	return EXIT_SUCCESS;
}
