/* Copyright (c) 2007
   Colin Dewey (University of Wisconsin-Madison)
   cdewey@biostat.wisc.edu
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "deweylab/bio/formats/sdb/DB.hh"
#include "deweylab/bio/formats/fasta.hh"
#include "deweylab/bio/alphabet/Nucleotide.hh"
#include "deweylab/bio/alphabet/AminoAcid.hh"
#include "deweylab/util/string.hh"
#include "deweylab/util/options.hh"
using namespace deweylab;

#include <iostream>
#include <stdexcept>
#include <string>

int main(int argc, const char* argv[]) {
	// Increase speed of input/output to standard streams
	std::ios::sync_with_stdio(false);

	// Initialize options to defaults
	bool number_title = false;
	bool unmask = false;
	bool hardmask = false;
	bool protein = false;
	std::string dbFilename;

	util::options::Parser parser("", "");
	parser.addStoreTrueOpt('t', "number-titles", 
						   "use record numbers as FASTA title in output",
						   number_title);
	parser.addStoreTrueOpt('u', "unmask",
						   "output all sequences in upper case (lose softmask formatting)",
						   unmask);
	parser.addStoreTrueOpt('h', "hardmask",
						   "hardmask the sequences.  Masked residues will be replaced by 'N' for DNA (default) and 'X' for protein",
						   hardmask);
	parser.addStoreTrueOpt('p', "protein",
						   "treat sequences as protein.  Only used in combination with the hardmask option",
						   protein);
	parser.addStoreArg("dbFile", "", dbFilename);
	parser.parse(argv, argv + argc);
	
	try {
		if (unmask and hardmask) {
			throw std::runtime_error("Cannot specify both unmask and "
									 "hardmask options");
		}
		
		bio::formats::sdb::DB db;
		db.open(dbFilename, true);

		bio::formats::fasta::OutputStream fastaOutStream(std::cout);
		bio::formats::fasta::Record rec;
		for (bio::formats::sdb::DB::Iterator it = db.begin(); it != db.end(); ++it) {
			rec.id = number_title ?
				util::string::toString(it->getRecNum()) : it->getTitle();
			rec.sequence = it->getSeq();
			if (unmask) {
				if (protein) {
					bio::alphabet::AminoAcid::unMaskInPlace(rec.sequence);
				} else {
					bio::alphabet::Nucleotide::unMaskInPlace(rec.sequence);
				}
			} else if (hardmask) {
				if (protein) {
					bio::alphabet::AminoAcid::hardMaskInPlace(rec.sequence);
				} else {
						bio::alphabet::Nucleotide::hardMaskInPlace(rec.sequence);
				}
			}
			fastaOutStream << rec;
		}
		
	} catch (const std::runtime_error& e) {
		std::cerr << "Error: " << e.what() << '\n';
		return EXIT_FAILURE;
	}	

	return EXIT_SUCCESS;
}
